/* arraylist.vala
 *
 * Copyright (C) 2004-2005  Novell, Inc
 * Copyright (C) 2005  David Waite
 * Copyright (C) 2007-2008  Jürg Billeter
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
 * Author:
 * 	Jürg Billeter <j@bitron.ch>
 */

#include <gee/arraylist.h>
#include <string.h>


#define GEE_ARRAY_LIST_TYPE_ITERATOR (gee_array_list_iterator_get_type ())
#define GEE_ARRAY_LIST_ITERATOR(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), GEE_ARRAY_LIST_TYPE_ITERATOR, GeeArrayListIterator))
#define GEE_ARRAY_LIST_ITERATOR_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), GEE_ARRAY_LIST_TYPE_ITERATOR, GeeArrayListIteratorClass))
#define GEE_ARRAY_LIST_IS_ITERATOR(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GEE_ARRAY_LIST_TYPE_ITERATOR))
#define GEE_ARRAY_LIST_IS_ITERATOR_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), GEE_ARRAY_LIST_TYPE_ITERATOR))
#define GEE_ARRAY_LIST_ITERATOR_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), GEE_ARRAY_LIST_TYPE_ITERATOR, GeeArrayListIteratorClass))

typedef struct _GeeArrayListIterator GeeArrayListIterator;
typedef struct _GeeArrayListIteratorClass GeeArrayListIteratorClass;
typedef struct _GeeArrayListIteratorPrivate GeeArrayListIteratorPrivate;

struct _GeeArrayListIterator {
	GeeCollectionObject parent_instance;
	GeeArrayListIteratorPrivate * priv;
	gint _stamp;
};

struct _GeeArrayListIteratorClass {
	GeeCollectionObjectClass parent_class;
};



struct _GeeArrayListPrivate {
	gpointer* _items;
	gint _items_length1;
	gint _items_size;
	gint _size;
	GEqualFunc _equal_func;
	gint _stamp;
	GType g_type;
	GBoxedCopyFunc g_dup_func;
	GDestroyNotify g_destroy_func;
};

#define GEE_ARRAY_LIST_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GEE_TYPE_ARRAY_LIST, GeeArrayListPrivate))
enum  {
	GEE_ARRAY_LIST_DUMMY_PROPERTY
};
static GType gee_array_list_real_get_element_type (GeeIterable* base);
static GeeIterator* gee_array_list_real_iterator (GeeIterable* base);
static gboolean gee_array_list_real_contains (GeeCollection* base, gconstpointer item);
static gint gee_array_list_real_index_of (GeeList* base, gconstpointer item);
static gpointer gee_array_list_real_get (GeeList* base, gint index);
static void gee_array_list_real_set (GeeList* base, gint index, gconstpointer item);
static gboolean gee_array_list_real_add (GeeCollection* base, gconstpointer item);
static void gee_array_list_real_insert (GeeList* base, gint index, gconstpointer item);
static gboolean gee_array_list_real_remove (GeeCollection* base, gconstpointer item);
static void gee_array_list_real_remove_at (GeeList* base, gint index);
static void gee_array_list_real_clear (GeeCollection* base);
static void gee_array_list_shift (GeeArrayList* self, gint start, gint delta);
static void gee_array_list_grow_if_needed (GeeArrayList* self, gint new_count);
static void gee_array_list_set_capacity (GeeArrayList* self, gint value);
struct _GeeArrayListIteratorPrivate {
	GeeArrayList* _list;
	gint _index;
	GType g_type;
	GBoxedCopyFunc g_dup_func;
	GDestroyNotify g_destroy_func;
};

#define GEE_ARRAY_LIST_ITERATOR_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GEE_ARRAY_LIST_TYPE_ITERATOR, GeeArrayListIteratorPrivate))
enum  {
	GEE_ARRAY_LIST_ITERATOR_DUMMY_PROPERTY
};
static GeeArrayListIterator* gee_array_list_iterator_construct (GType object_type, GType g_type, GBoxedCopyFunc g_dup_func, GDestroyNotify g_destroy_func, GeeArrayList* list);
static GeeArrayListIterator* gee_array_list_iterator_new (GType g_type, GBoxedCopyFunc g_dup_func, GDestroyNotify g_destroy_func, GeeArrayList* list);
static gboolean gee_array_list_iterator_real_next (GeeIterator* base);
static gpointer gee_array_list_iterator_real_get (GeeIterator* base);
static void gee_array_list_iterator_set_list (GeeArrayListIterator* self, GeeArrayList* value);
static gpointer gee_array_list_iterator_parent_class = NULL;
static GeeIteratorIface* gee_array_list_iterator_gee_iterator_parent_iface = NULL;
static void gee_array_list_iterator_finalize (GeeCollectionObject* obj);
static GType gee_array_list_iterator_get_type (void);
static gpointer gee_array_list_parent_class = NULL;
static GeeIterableIface* gee_array_list_gee_iterable_parent_iface = NULL;
static GeeCollectionIface* gee_array_list_gee_collection_parent_iface = NULL;
static GeeListIface* gee_array_list_gee_list_parent_iface = NULL;
static void gee_array_list_finalize (GeeCollectionObject* obj);
static void _vala_array_free (gpointer array, gint array_length, GDestroyNotify destroy_func);
static void _vala_array_move (gpointer array, gsize element_size, gint src, gint dest, gint length);



GeeArrayList* gee_array_list_construct (GType object_type, GType g_type, GBoxedCopyFunc g_dup_func, GDestroyNotify g_destroy_func, GEqualFunc equal_func) {
	GeeArrayList* self;
	self = (GeeArrayList*) g_type_create_instance (object_type);
	self->priv->g_type = g_type;
	self->priv->g_dup_func = g_dup_func;
	self->priv->g_destroy_func = g_destroy_func;
	gee_array_list_set_equal_func (self, equal_func);
	return self;
}


GeeArrayList* gee_array_list_new (GType g_type, GBoxedCopyFunc g_dup_func, GDestroyNotify g_destroy_func, GEqualFunc equal_func) {
	return gee_array_list_construct (GEE_TYPE_ARRAY_LIST, g_type, g_dup_func, g_destroy_func, equal_func);
}


static GType gee_array_list_real_get_element_type (GeeIterable* base) {
	GeeArrayList * self;
	self = (GeeArrayList*) base;
	return self->priv->g_type;
}


static GeeIterator* gee_array_list_real_iterator (GeeIterable* base) {
	GeeArrayList * self;
	self = (GeeArrayList*) base;
	return (GeeIterator*) gee_array_list_iterator_new (self->priv->g_type, (GBoxedCopyFunc) self->priv->g_dup_func, self->priv->g_destroy_func, self);
}


static gboolean gee_array_list_real_contains (GeeCollection* base, gconstpointer item) {
	GeeArrayList * self;
	self = (GeeArrayList*) base;
	return gee_list_index_of ((GeeList*) self, item) != (-1);
}


static gint gee_array_list_real_index_of (GeeList* base, gconstpointer item) {
	GeeArrayList * self;
	self = (GeeArrayList*) base;
	{
		gint index;
		index = 0;
		for (; index < self->priv->_size; index++) {
			if (self->priv->_equal_func (self->priv->_items[index], item)) {
				return index;
			}
		}
	}
	return -1;
}


static gpointer gee_array_list_real_get (GeeList* base, gint index) {
	GeeArrayList * self;
	gboolean _tmp0;
	gconstpointer _tmp1;
	self = (GeeArrayList*) base;
	_tmp0 = FALSE;
	if (index >= 0) {
		_tmp0 = index < self->priv->_size;
	} else {
		_tmp0 = FALSE;
	}
	g_assert (_tmp0);
	_tmp1 = NULL;
	return (_tmp1 = self->priv->_items[index], ((_tmp1 == NULL) || (self->priv->g_dup_func == NULL)) ? ((gpointer) _tmp1) : self->priv->g_dup_func ((gpointer) _tmp1));
}


static void gee_array_list_real_set (GeeList* base, gint index, gconstpointer item) {
	GeeArrayList * self;
	gboolean _tmp0;
	gpointer _tmp2;
	gconstpointer _tmp1;
	self = (GeeArrayList*) base;
	_tmp0 = FALSE;
	if (index >= 0) {
		_tmp0 = index < self->priv->_size;
	} else {
		_tmp0 = FALSE;
	}
	g_assert (_tmp0);
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_items[index] = (_tmp2 = (_tmp1 = item, ((_tmp1 == NULL) || (self->priv->g_dup_func == NULL)) ? ((gpointer) _tmp1) : self->priv->g_dup_func ((gpointer) _tmp1)), ((self->priv->_items[index] == NULL) || (self->priv->g_destroy_func == NULL)) ? NULL : (self->priv->_items[index] = (self->priv->g_destroy_func (self->priv->_items[index]), NULL)), _tmp2);
}


static gboolean gee_array_list_real_add (GeeCollection* base, gconstpointer item) {
	GeeArrayList * self;
	gpointer _tmp2;
	gpointer *_tmp1;
	gconstpointer _tmp0;
	self = (GeeArrayList*) base;
	if (self->priv->_size == self->priv->_items_length1) {
		gee_array_list_grow_if_needed (self, 1);
	}
	_tmp2 = NULL;
	_tmp0 = NULL;
	_tmp1 = &self->priv->_items[self->priv->_size++];
	(*_tmp1) = (_tmp2 = (_tmp0 = item, ((_tmp0 == NULL) || (self->priv->g_dup_func == NULL)) ? ((gpointer) _tmp0) : self->priv->g_dup_func ((gpointer) _tmp0)), (((*_tmp1) == NULL) || (self->priv->g_destroy_func == NULL)) ? NULL : ((*_tmp1) = (self->priv->g_destroy_func ((*_tmp1)), NULL)), _tmp2);
	self->priv->_stamp++;
	return TRUE;
}


static void gee_array_list_real_insert (GeeList* base, gint index, gconstpointer item) {
	GeeArrayList * self;
	gboolean _tmp0;
	gpointer _tmp2;
	gconstpointer _tmp1;
	self = (GeeArrayList*) base;
	_tmp0 = FALSE;
	if (index >= 0) {
		_tmp0 = index <= self->priv->_size;
	} else {
		_tmp0 = FALSE;
	}
	g_assert (_tmp0);
	if (self->priv->_size == self->priv->_items_length1) {
		gee_array_list_grow_if_needed (self, 1);
	}
	gee_array_list_shift (self, index, 1);
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_items[index] = (_tmp2 = (_tmp1 = item, ((_tmp1 == NULL) || (self->priv->g_dup_func == NULL)) ? ((gpointer) _tmp1) : self->priv->g_dup_func ((gpointer) _tmp1)), ((self->priv->_items[index] == NULL) || (self->priv->g_destroy_func == NULL)) ? NULL : (self->priv->_items[index] = (self->priv->g_destroy_func (self->priv->_items[index]), NULL)), _tmp2);
	self->priv->_stamp++;
}


static gboolean gee_array_list_real_remove (GeeCollection* base, gconstpointer item) {
	GeeArrayList * self;
	self = (GeeArrayList*) base;
	{
		gint index;
		index = 0;
		for (; index < self->priv->_size; index++) {
			if (self->priv->_equal_func (self->priv->_items[index], item)) {
				gee_list_remove_at ((GeeList*) self, index);
				return TRUE;
			}
		}
	}
	return FALSE;
}


static void gee_array_list_real_remove_at (GeeList* base, gint index) {
	GeeArrayList * self;
	gboolean _tmp0;
	gpointer _tmp1;
	self = (GeeArrayList*) base;
	_tmp0 = FALSE;
	if (index >= 0) {
		_tmp0 = index < self->priv->_size;
	} else {
		_tmp0 = FALSE;
	}
	g_assert (_tmp0);
	_tmp1 = NULL;
	self->priv->_items[index] = (_tmp1 = NULL, ((self->priv->_items[index] == NULL) || (self->priv->g_destroy_func == NULL)) ? NULL : (self->priv->_items[index] = (self->priv->g_destroy_func (self->priv->_items[index]), NULL)), _tmp1);
	gee_array_list_shift (self, index + 1, -1);
	self->priv->_stamp++;
}


static void gee_array_list_real_clear (GeeCollection* base) {
	GeeArrayList * self;
	self = (GeeArrayList*) base;
	{
		gint index;
		index = 0;
		for (; index < self->priv->_size; index++) {
			gpointer _tmp0;
			_tmp0 = NULL;
			self->priv->_items[index] = (_tmp0 = NULL, ((self->priv->_items[index] == NULL) || (self->priv->g_destroy_func == NULL)) ? NULL : (self->priv->_items[index] = (self->priv->g_destroy_func (self->priv->_items[index]), NULL)), _tmp0);
		}
	}
	self->priv->_size = 0;
	self->priv->_stamp++;
}


static void gee_array_list_shift (GeeArrayList* self, gint start, gint delta) {
	gboolean _tmp0;
	gboolean _tmp1;
	g_return_if_fail (self != NULL);
	_tmp0 = FALSE;
	_tmp1 = FALSE;
	if (start >= 0) {
		_tmp1 = start <= self->priv->_size;
	} else {
		_tmp1 = FALSE;
	}
	if (_tmp1) {
		_tmp0 = start >= (-delta);
	} else {
		_tmp0 = FALSE;
	}
	g_assert (_tmp0);
	_vala_array_move (self->priv->_items, sizeof (gpointer), start, start + delta, self->priv->_size - start);
	self->priv->_size = self->priv->_size + delta;
}


static void gee_array_list_grow_if_needed (GeeArrayList* self, gint new_count) {
	gint minimum_size;
	g_return_if_fail (self != NULL);
	g_assert (new_count >= 0);
	minimum_size = self->priv->_size + new_count;
	if (minimum_size > self->priv->_items_length1) {
		gint _tmp0;
		_tmp0 = 0;
		if (new_count > self->priv->_items_length1) {
			_tmp0 = minimum_size;
		} else {
			_tmp0 = 2 * self->priv->_items_length1;
		}
		/* double the capacity unless we add even more items at this time*/
		gee_array_list_set_capacity (self, _tmp0);
	}
}


static void gee_array_list_set_capacity (GeeArrayList* self, gint value) {
	gint _tmp0;
	g_return_if_fail (self != NULL);
	g_assert (value >= self->priv->_size);
	_tmp0 = value;
	self->priv->_items = g_renew (gpointer, self->priv->_items, value);
	(_tmp0 > self->priv->_items_length1) ? memset (self->priv->_items + self->priv->_items_length1, 0, sizeof (gpointer) * (_tmp0 - self->priv->_items_length1)) : NULL;
	self->priv->_items_length1 = _tmp0;
}


static gint gee_array_list_real_get_size (GeeCollection* base) {
	GeeArrayList* self;
	self = (GeeArrayList*) base;
	return self->priv->_size;
}


void gee_array_list_set_equal_func (GeeArrayList* self, GEqualFunc value) {
	g_return_if_fail (self != NULL);
	self->priv->_equal_func = value;
}


static GeeArrayListIterator* gee_array_list_iterator_construct (GType object_type, GType g_type, GBoxedCopyFunc g_dup_func, GDestroyNotify g_destroy_func, GeeArrayList* list) {
	GeeArrayListIterator* self;
	g_return_val_if_fail (list != NULL, NULL);
	self = (GeeArrayListIterator*) g_type_create_instance (object_type);
	self->priv->g_type = g_type;
	self->priv->g_dup_func = g_dup_func;
	self->priv->g_destroy_func = g_destroy_func;
	gee_array_list_iterator_set_list (self, list);
	return self;
}


static GeeArrayListIterator* gee_array_list_iterator_new (GType g_type, GBoxedCopyFunc g_dup_func, GDestroyNotify g_destroy_func, GeeArrayList* list) {
	return gee_array_list_iterator_construct (GEE_ARRAY_LIST_TYPE_ITERATOR, g_type, g_dup_func, g_destroy_func, list);
}


static gboolean gee_array_list_iterator_real_next (GeeIterator* base) {
	GeeArrayListIterator * self;
	self = (GeeArrayListIterator*) base;
	g_assert (self->_stamp == self->priv->_list->priv->_stamp);
	if (self->priv->_index < self->priv->_list->priv->_size) {
		self->priv->_index++;
	}
	return self->priv->_index < self->priv->_list->priv->_size;
}


static gpointer gee_array_list_iterator_real_get (GeeIterator* base) {
	GeeArrayListIterator * self;
	gboolean _tmp0;
	self = (GeeArrayListIterator*) base;
	g_assert (self->_stamp == self->priv->_list->priv->_stamp);
	_tmp0 = FALSE;
	if (self->priv->_index < 0) {
		_tmp0 = TRUE;
	} else {
		_tmp0 = self->priv->_index >= self->priv->_list->priv->_size;
	}
	if (_tmp0) {
		return NULL;
	}
	return gee_list_get ((GeeList*) self->priv->_list, self->priv->_index);
}


static void gee_array_list_iterator_set_list (GeeArrayListIterator* self, GeeArrayList* value) {
	GeeArrayList* _tmp1;
	GeeArrayList* _tmp0;
	g_return_if_fail (self != NULL);
	_tmp1 = NULL;
	_tmp0 = NULL;
	self->priv->_list = (_tmp1 = (_tmp0 = value, (_tmp0 == NULL) ? NULL : gee_collection_object_ref (_tmp0)), (self->priv->_list == NULL) ? NULL : (self->priv->_list = (gee_collection_object_unref (self->priv->_list), NULL)), _tmp1);
	self->_stamp = self->priv->_list->priv->_stamp;
}


static void gee_array_list_iterator_class_init (GeeArrayListIteratorClass * klass) {
	gee_array_list_iterator_parent_class = g_type_class_peek_parent (klass);
	GEE_COLLECTION_OBJECT_CLASS (klass)->finalize = gee_array_list_iterator_finalize;
	g_type_class_add_private (klass, sizeof (GeeArrayListIteratorPrivate));
}


static void gee_array_list_iterator_gee_iterator_interface_init (GeeIteratorIface * iface) {
	gee_array_list_iterator_gee_iterator_parent_iface = g_type_interface_peek_parent (iface);
	iface->next = gee_array_list_iterator_real_next;
	iface->get = gee_array_list_iterator_real_get;
}


static void gee_array_list_iterator_instance_init (GeeArrayListIterator * self) {
	self->priv = GEE_ARRAY_LIST_ITERATOR_GET_PRIVATE (self);
	self->priv->_index = -1;
	self->_stamp = 0;
}


static void gee_array_list_iterator_finalize (GeeCollectionObject* obj) {
	GeeArrayListIterator * self;
	self = GEE_ARRAY_LIST_ITERATOR (obj);
	(self->priv->_list == NULL) ? NULL : (self->priv->_list = (gee_collection_object_unref (self->priv->_list), NULL));
	GEE_COLLECTION_OBJECT_CLASS (gee_array_list_iterator_parent_class)->finalize (obj);
}


static GType gee_array_list_iterator_get_type (void) {
	static GType gee_array_list_iterator_type_id = 0;
	if (gee_array_list_iterator_type_id == 0) {
		static const GTypeInfo g_define_type_info = { sizeof (GeeArrayListIteratorClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) gee_array_list_iterator_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (GeeArrayListIterator), 0, (GInstanceInitFunc) gee_array_list_iterator_instance_init, NULL };
		static const GInterfaceInfo gee_iterator_info = { (GInterfaceInitFunc) gee_array_list_iterator_gee_iterator_interface_init, (GInterfaceFinalizeFunc) NULL, NULL};
		gee_array_list_iterator_type_id = g_type_register_static (GEE_TYPE_COLLECTION_OBJECT, "GeeArrayListIterator", &g_define_type_info, 0);
		g_type_add_interface_static (gee_array_list_iterator_type_id, GEE_TYPE_ITERATOR, &gee_iterator_info);
	}
	return gee_array_list_iterator_type_id;
}


static void gee_array_list_class_init (GeeArrayListClass * klass) {
	gee_array_list_parent_class = g_type_class_peek_parent (klass);
	GEE_COLLECTION_OBJECT_CLASS (klass)->finalize = gee_array_list_finalize;
	g_type_class_add_private (klass, sizeof (GeeArrayListPrivate));
}


static void gee_array_list_gee_iterable_interface_init (GeeIterableIface * iface) {
	gee_array_list_gee_iterable_parent_iface = g_type_interface_peek_parent (iface);
	iface->get_element_type = gee_array_list_real_get_element_type;
	iface->iterator = gee_array_list_real_iterator;
}


static void gee_array_list_gee_collection_interface_init (GeeCollectionIface * iface) {
	gee_array_list_gee_collection_parent_iface = g_type_interface_peek_parent (iface);
	iface->contains = gee_array_list_real_contains;
	iface->add = gee_array_list_real_add;
	iface->remove = gee_array_list_real_remove;
	iface->clear = gee_array_list_real_clear;
	iface->get_size = gee_array_list_real_get_size;
}


static void gee_array_list_gee_list_interface_init (GeeListIface * iface) {
	gee_array_list_gee_list_parent_iface = g_type_interface_peek_parent (iface);
	iface->index_of = gee_array_list_real_index_of;
	iface->get = gee_array_list_real_get;
	iface->set = gee_array_list_real_set;
	iface->insert = gee_array_list_real_insert;
	iface->remove_at = gee_array_list_real_remove_at;
}


static void gee_array_list_instance_init (GeeArrayList * self) {
	self->priv = GEE_ARRAY_LIST_GET_PRIVATE (self);
	self->priv->_items = g_new0 (gpointer, 4);
	self->priv->_items_length1 = 4;
	self->priv->_stamp = 0;
}


static void gee_array_list_finalize (GeeCollectionObject* obj) {
	GeeArrayList * self;
	self = GEE_ARRAY_LIST (obj);
	self->priv->_items = (_vala_array_free (self->priv->_items, self->priv->_items_length1, (GDestroyNotify) self->priv->g_destroy_func), NULL);
	GEE_COLLECTION_OBJECT_CLASS (gee_array_list_parent_class)->finalize (obj);
}


GType gee_array_list_get_type (void) {
	static GType gee_array_list_type_id = 0;
	if (gee_array_list_type_id == 0) {
		static const GTypeInfo g_define_type_info = { sizeof (GeeArrayListClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) gee_array_list_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (GeeArrayList), 0, (GInstanceInitFunc) gee_array_list_instance_init, NULL };
		static const GInterfaceInfo gee_iterable_info = { (GInterfaceInitFunc) gee_array_list_gee_iterable_interface_init, (GInterfaceFinalizeFunc) NULL, NULL};
		static const GInterfaceInfo gee_collection_info = { (GInterfaceInitFunc) gee_array_list_gee_collection_interface_init, (GInterfaceFinalizeFunc) NULL, NULL};
		static const GInterfaceInfo gee_list_info = { (GInterfaceInitFunc) gee_array_list_gee_list_interface_init, (GInterfaceFinalizeFunc) NULL, NULL};
		gee_array_list_type_id = g_type_register_static (GEE_TYPE_COLLECTION_OBJECT, "GeeArrayList", &g_define_type_info, 0);
		g_type_add_interface_static (gee_array_list_type_id, GEE_TYPE_ITERABLE, &gee_iterable_info);
		g_type_add_interface_static (gee_array_list_type_id, GEE_TYPE_COLLECTION, &gee_collection_info);
		g_type_add_interface_static (gee_array_list_type_id, GEE_TYPE_LIST, &gee_list_info);
	}
	return gee_array_list_type_id;
}


static void _vala_array_free (gpointer array, gint array_length, GDestroyNotify destroy_func) {
	if ((array != NULL) && (destroy_func != NULL)) {
		int i;
		for (i = 0; i < array_length; i = i + 1) {
			if (((gpointer*) array)[i] != NULL) {
				destroy_func (((gpointer*) array)[i]);
			}
		}
	}
	g_free (array);
}


static void _vala_array_move (gpointer array, gsize element_size, gint src, gint dest, gint length) {
	g_memmove (((char*) array) + (dest * element_size), ((char*) array) + (src * element_size), length * element_size);
	if (src < dest) {
		memset (((char*) array) + (src * element_size), 0, (dest - src) * element_size);
	} else {
		memset (((char*) array) + ((dest + length) * element_size), 0, (src - dest) * element_size);
	}
}




