/* valacatchvala
 *
 * Copyright (C) 2007-2008  Jürg Billeter
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
 * Author:
 * 	Jürg Billeter <j@bitron.ch>
 */

#include <vala/valacatchclause.h>
#include <vala/valablock.h>
#include <vala/valalocalvariable.h>
#include <vala/valadatatype.h>
#include <vala/valasourcereference.h>
#include <vala/valacodevisitor.h>
#include <vala/valasemanticanalyzer.h>
#include <vala/valaexpression.h>
#include <vala/valasymbol.h>
#include <vala/valascope.h>
#include <vala/valaerrortype.h>
#include <vala/valaerrordomain.h>
#include <vala/valaerrorcode.h>
#include <vala/valasourcefile.h>




struct _ValaCatchClausePrivate {
	char* _variable_name;
	ValaBlock* _body;
	ValaLocalVariable* _error_variable;
	char* _clabel_name;
	ValaDataType* _data_type;
};

#define VALA_CATCH_CLAUSE_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), VALA_TYPE_CATCH_CLAUSE, ValaCatchClausePrivate))
enum  {
	VALA_CATCH_CLAUSE_DUMMY_PROPERTY
};
static void vala_catch_clause_real_accept (ValaCodeNode* base, ValaCodeVisitor* visitor);
static void vala_catch_clause_real_accept_children (ValaCodeNode* base, ValaCodeVisitor* visitor);
static void vala_catch_clause_real_replace_type (ValaCodeNode* base, ValaDataType* old_type, ValaDataType* new_type);
static gboolean vala_catch_clause_real_check (ValaCodeNode* base, ValaSemanticAnalyzer* analyzer);
static void vala_catch_clause_real_get_defined_variables (ValaCodeNode* base, GeeCollection* collection);
static gpointer vala_catch_clause_parent_class = NULL;
static void vala_catch_clause_finalize (ValaCodeNode* obj);



/**
 * Creates a new catch 
 *
 * @param type_reference   error type
 * @param variable_name    error variable name
 * @param body             error handler body
 * @param source_reference reference to source code
 * @return                 newly created catch clause
 */
ValaCatchClause* vala_catch_clause_construct (GType object_type, ValaDataType* error_type, const char* variable_name, ValaBlock* body, ValaSourceReference* source_reference) {
	ValaCatchClause* self;
	g_return_val_if_fail (body != NULL, NULL);
	self = (ValaCatchClause*) g_type_create_instance (object_type);
	vala_catch_clause_set_error_type (self, error_type);
	vala_catch_clause_set_variable_name (self, variable_name);
	vala_catch_clause_set_body (self, body);
	vala_code_node_set_source_reference ((ValaCodeNode*) self, source_reference);
	return self;
}


ValaCatchClause* vala_catch_clause_new (ValaDataType* error_type, const char* variable_name, ValaBlock* body, ValaSourceReference* source_reference) {
	return vala_catch_clause_construct (VALA_TYPE_CATCH_CLAUSE, error_type, variable_name, body, source_reference);
}


static void vala_catch_clause_real_accept (ValaCodeNode* base, ValaCodeVisitor* visitor) {
	ValaCatchClause * self;
	self = (ValaCatchClause*) base;
	g_return_if_fail (visitor != NULL);
	vala_code_visitor_visit_catch_clause (visitor, self);
}


static void vala_catch_clause_real_accept_children (ValaCodeNode* base, ValaCodeVisitor* visitor) {
	ValaCatchClause * self;
	self = (ValaCatchClause*) base;
	g_return_if_fail (visitor != NULL);
	if (vala_catch_clause_get_error_type (self) != NULL) {
		vala_code_node_accept ((ValaCodeNode*) vala_catch_clause_get_error_type (self), visitor);
	}
	vala_code_node_accept ((ValaCodeNode*) self->priv->_body, visitor);
}


static void vala_catch_clause_real_replace_type (ValaCodeNode* base, ValaDataType* old_type, ValaDataType* new_type) {
	ValaCatchClause * self;
	self = (ValaCatchClause*) base;
	g_return_if_fail (old_type != NULL);
	g_return_if_fail (new_type != NULL);
	if (vala_catch_clause_get_error_type (self) == old_type) {
		vala_catch_clause_set_error_type (self, new_type);
	}
}


static gboolean vala_catch_clause_real_check (ValaCodeNode* base, ValaSemanticAnalyzer* analyzer) {
	ValaCatchClause * self;
	self = (ValaCatchClause*) base;
	g_return_val_if_fail (analyzer != NULL, FALSE);
	if (vala_code_node_get_checked ((ValaCodeNode*) self)) {
		return !vala_code_node_get_error ((ValaCodeNode*) self);
	}
	vala_code_node_set_checked ((ValaCodeNode*) self, TRUE);
	if (vala_catch_clause_get_error_type (self) != NULL) {
		ValaLocalVariable* _tmp2;
		ValaDataType* _tmp1;
		_tmp2 = NULL;
		_tmp1 = NULL;
		vala_catch_clause_set_error_variable (self, _tmp2 = vala_local_variable_new (_tmp1 = vala_data_type_copy (vala_catch_clause_get_error_type (self)), self->priv->_variable_name, NULL, NULL));
		(_tmp2 == NULL) ? NULL : (_tmp2 = (vala_code_node_unref (_tmp2), NULL));
		(_tmp1 == NULL) ? NULL : (_tmp1 = (vala_code_node_unref (_tmp1), NULL));
		vala_scope_add (vala_symbol_get_scope ((ValaSymbol*) self->priv->_body), self->priv->_variable_name, (ValaSymbol*) self->priv->_error_variable);
		vala_block_add_local_variable (self->priv->_body, self->priv->_error_variable);
		vala_code_node_set_checked ((ValaCodeNode*) self->priv->_error_variable, TRUE);
	} else {
		ValaErrorType* _tmp3;
		_tmp3 = NULL;
		vala_catch_clause_set_error_type (self, (ValaDataType*) (_tmp3 = vala_error_type_new (NULL, NULL, vala_code_node_get_source_reference ((ValaCodeNode*) self))));
		(_tmp3 == NULL) ? NULL : (_tmp3 = (vala_code_node_unref (_tmp3), NULL));
	}
	vala_code_node_check ((ValaCodeNode*) vala_catch_clause_get_error_type (self), analyzer);
	vala_source_file_add_type_dependency (vala_semantic_analyzer_get_current_source_file (analyzer), vala_catch_clause_get_error_type (self), VALA_SOURCE_FILE_DEPENDENCY_TYPE_SOURCE);
	vala_code_node_check ((ValaCodeNode*) self->priv->_body, analyzer);
	return !vala_code_node_get_error ((ValaCodeNode*) self);
}


static void vala_catch_clause_real_get_defined_variables (ValaCodeNode* base, GeeCollection* collection) {
	ValaCatchClause * self;
	self = (ValaCatchClause*) base;
	g_return_if_fail (collection != NULL);
	if (self->priv->_error_variable != NULL) {
		gee_collection_add (collection, self->priv->_error_variable);
	}
}


ValaDataType* vala_catch_clause_get_error_type (ValaCatchClause* self) {
	g_return_val_if_fail (self != NULL, NULL);
	return self->priv->_data_type;
}


void vala_catch_clause_set_error_type (ValaCatchClause* self, ValaDataType* value) {
	ValaDataType* _tmp2;
	ValaDataType* _tmp1;
	g_return_if_fail (self != NULL);
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_data_type = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL) ? NULL : vala_code_node_ref (_tmp1)), (self->priv->_data_type == NULL) ? NULL : (self->priv->_data_type = (vala_code_node_unref (self->priv->_data_type), NULL)), _tmp2);
	if (self->priv->_data_type != NULL) {
		vala_code_node_set_parent_node ((ValaCodeNode*) self->priv->_data_type, (ValaCodeNode*) self);
	}
}


const char* vala_catch_clause_get_variable_name (ValaCatchClause* self) {
	g_return_val_if_fail (self != NULL, NULL);
	return self->priv->_variable_name;
}


void vala_catch_clause_set_variable_name (ValaCatchClause* self, const char* value) {
	char* _tmp2;
	const char* _tmp1;
	g_return_if_fail (self != NULL);
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_variable_name = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL) ? NULL : g_strdup (_tmp1)), self->priv->_variable_name = (g_free (self->priv->_variable_name), NULL), _tmp2);
}


ValaBlock* vala_catch_clause_get_body (ValaCatchClause* self) {
	g_return_val_if_fail (self != NULL, NULL);
	return self->priv->_body;
}


void vala_catch_clause_set_body (ValaCatchClause* self, ValaBlock* value) {
	ValaBlock* _tmp2;
	ValaBlock* _tmp1;
	g_return_if_fail (self != NULL);
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_body = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL) ? NULL : vala_code_node_ref (_tmp1)), (self->priv->_body == NULL) ? NULL : (self->priv->_body = (vala_code_node_unref (self->priv->_body), NULL)), _tmp2);
}


ValaLocalVariable* vala_catch_clause_get_error_variable (ValaCatchClause* self) {
	g_return_val_if_fail (self != NULL, NULL);
	return self->priv->_error_variable;
}


void vala_catch_clause_set_error_variable (ValaCatchClause* self, ValaLocalVariable* value) {
	ValaLocalVariable* _tmp2;
	ValaLocalVariable* _tmp1;
	g_return_if_fail (self != NULL);
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_error_variable = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL) ? NULL : vala_code_node_ref (_tmp1)), (self->priv->_error_variable == NULL) ? NULL : (self->priv->_error_variable = (vala_code_node_unref (self->priv->_error_variable), NULL)), _tmp2);
}


const char* vala_catch_clause_get_clabel_name (ValaCatchClause* self) {
	g_return_val_if_fail (self != NULL, NULL);
	return self->priv->_clabel_name;
}


void vala_catch_clause_set_clabel_name (ValaCatchClause* self, const char* value) {
	char* _tmp2;
	const char* _tmp1;
	g_return_if_fail (self != NULL);
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_clabel_name = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL) ? NULL : g_strdup (_tmp1)), self->priv->_clabel_name = (g_free (self->priv->_clabel_name), NULL), _tmp2);
}


static void vala_catch_clause_class_init (ValaCatchClauseClass * klass) {
	vala_catch_clause_parent_class = g_type_class_peek_parent (klass);
	VALA_CODE_NODE_CLASS (klass)->finalize = vala_catch_clause_finalize;
	g_type_class_add_private (klass, sizeof (ValaCatchClausePrivate));
	VALA_CODE_NODE_CLASS (klass)->accept = vala_catch_clause_real_accept;
	VALA_CODE_NODE_CLASS (klass)->accept_children = vala_catch_clause_real_accept_children;
	VALA_CODE_NODE_CLASS (klass)->replace_type = vala_catch_clause_real_replace_type;
	VALA_CODE_NODE_CLASS (klass)->check = vala_catch_clause_real_check;
	VALA_CODE_NODE_CLASS (klass)->get_defined_variables = vala_catch_clause_real_get_defined_variables;
}


static void vala_catch_clause_instance_init (ValaCatchClause * self) {
	self->priv = VALA_CATCH_CLAUSE_GET_PRIVATE (self);
}


static void vala_catch_clause_finalize (ValaCodeNode* obj) {
	ValaCatchClause * self;
	self = VALA_CATCH_CLAUSE (obj);
	self->priv->_variable_name = (g_free (self->priv->_variable_name), NULL);
	(self->priv->_body == NULL) ? NULL : (self->priv->_body = (vala_code_node_unref (self->priv->_body), NULL));
	(self->priv->_error_variable == NULL) ? NULL : (self->priv->_error_variable = (vala_code_node_unref (self->priv->_error_variable), NULL));
	self->priv->_clabel_name = (g_free (self->priv->_clabel_name), NULL);
	(self->priv->_data_type == NULL) ? NULL : (self->priv->_data_type = (vala_code_node_unref (self->priv->_data_type), NULL));
	VALA_CODE_NODE_CLASS (vala_catch_clause_parent_class)->finalize (obj);
}


GType vala_catch_clause_get_type (void) {
	static GType vala_catch_clause_type_id = 0;
	if (vala_catch_clause_type_id == 0) {
		static const GTypeInfo g_define_type_info = { sizeof (ValaCatchClauseClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) vala_catch_clause_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (ValaCatchClause), 0, (GInstanceInitFunc) vala_catch_clause_instance_init, NULL };
		vala_catch_clause_type_id = g_type_register_static (VALA_TYPE_CODE_NODE, "ValaCatchClause", &g_define_type_info, 0);
	}
	return vala_catch_clause_type_id;
}




