/* valascanner.vala
 *
 * Copyright (C) 2008-2009  Jürg Billeter
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
 * Author:
 * 	Jürg Billeter <j@bitron.ch>
 */

#include <vala/valascanner.h>
#include <vala/valareport.h>
#include <vala/valasourcereference.h>
#include <gobject/gvaluecollector.h>




struct _ValaScannerPrivate {
	ValaSourceFile* _source_file;
	gchar* current;
	gchar* end;
	gint line;
	gint column;
	char* _comment;
};

#define VALA_SCANNER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), VALA_TYPE_SCANNER, ValaScannerPrivate))
enum  {
	VALA_SCANNER_DUMMY_PROPERTY
};
static gboolean vala_scanner_is_ident_char (ValaScanner* self, gchar c);
static ValaTokenType vala_scanner_get_identifier_or_keyword (ValaScanner* self, gchar* begin, gint len);
static ValaTokenType vala_scanner_read_number (ValaScanner* self);
static gboolean vala_scanner_matches (ValaScanner* self, gchar* begin, const char* keyword);
static gboolean vala_scanner_whitespace (ValaScanner* self);
static gboolean vala_scanner_comment (ValaScanner* self);
static void vala_scanner_space (ValaScanner* self);
static void vala_scanner_push_comment (ValaScanner* self, const char* comment_item, gboolean file_comment);
static void vala_scanner_set_source_file (ValaScanner* self, ValaSourceFile* value);
static gpointer vala_scanner_parent_class = NULL;
static void vala_scanner_finalize (ValaScanner* obj);



ValaScanner* vala_scanner_construct (GType object_type, ValaSourceFile* source_file) {
	ValaScanner* self;
	gchar* begin;
	g_return_val_if_fail (source_file != NULL, NULL);
	self = (ValaScanner*) g_type_create_instance (object_type);
	vala_scanner_set_source_file (self, source_file);
	begin = vala_source_file_get_mapped_contents (source_file);
	self->priv->end = begin + vala_source_file_get_mapped_length (source_file);
	self->priv->current = begin;
	self->priv->line = 1;
	self->priv->column = 1;
	return self;
}


ValaScanner* vala_scanner_new (ValaSourceFile* source_file) {
	return vala_scanner_construct (VALA_TYPE_SCANNER, source_file);
}


static gboolean vala_scanner_is_ident_char (ValaScanner* self, gchar c) {
	gboolean _tmp0;
	g_return_val_if_fail (self != NULL, FALSE);
	_tmp0 = FALSE;
	if (g_ascii_isalnum (c)) {
		_tmp0 = TRUE;
	} else {
		_tmp0 = c == '_';
	}
	return _tmp0;
}


static ValaTokenType vala_scanner_get_identifier_or_keyword (ValaScanner* self, gchar* begin, gint len) {
	g_return_val_if_fail (self != NULL, 0);
	switch (len) {
		case 2:
		{
			switch (begin[0]) {
				case 'a':
				{
					if (vala_scanner_matches (self, begin, "as")) {
						return VALA_TOKEN_TYPE_AS;
					}
					break;
				}
				case 'd':
				{
					if (vala_scanner_matches (self, begin, "do")) {
						return VALA_TOKEN_TYPE_DO;
					}
					break;
				}
				case 'i':
				{
					switch (begin[1]) {
						case 'f':
						{
							return VALA_TOKEN_TYPE_IF;
						}
						case 'n':
						{
							return VALA_TOKEN_TYPE_IN;
						}
						case 's':
						{
							return VALA_TOKEN_TYPE_IS;
						}
					}
					break;
				}
			}
			break;
		}
		case 3:
		{
			switch (begin[0]) {
				case 'f':
				{
					if (vala_scanner_matches (self, begin, "for")) {
						return VALA_TOKEN_TYPE_FOR;
					}
					break;
				}
				case 'g':
				{
					if (vala_scanner_matches (self, begin, "get")) {
						return VALA_TOKEN_TYPE_GET;
					}
					break;
				}
				case 'n':
				{
					if (vala_scanner_matches (self, begin, "new")) {
						return VALA_TOKEN_TYPE_NEW;
					}
					break;
				}
				case 'o':
				{
					if (vala_scanner_matches (self, begin, "out")) {
						return VALA_TOKEN_TYPE_OUT;
					}
					break;
				}
				case 'r':
				{
					if (vala_scanner_matches (self, begin, "ref")) {
						return VALA_TOKEN_TYPE_REF;
					}
					break;
				}
				case 's':
				{
					if (vala_scanner_matches (self, begin, "set")) {
						return VALA_TOKEN_TYPE_SET;
					}
					break;
				}
				case 't':
				{
					if (vala_scanner_matches (self, begin, "try")) {
						return VALA_TOKEN_TYPE_TRY;
					}
					break;
				}
				case 'v':
				{
					if (vala_scanner_matches (self, begin, "var")) {
						return VALA_TOKEN_TYPE_VAR;
					}
					break;
				}
			}
			break;
		}
		case 4:
		{
			switch (begin[0]) {
				case 'b':
				{
					if (vala_scanner_matches (self, begin, "base")) {
						return VALA_TOKEN_TYPE_BASE;
					}
					break;
				}
				case 'c':
				{
					if (vala_scanner_matches (self, begin, "case")) {
						return VALA_TOKEN_TYPE_CASE;
					}
					break;
				}
				case 'e':
				{
					switch (begin[1]) {
						case 'l':
						{
							if (vala_scanner_matches (self, begin, "else")) {
								return VALA_TOKEN_TYPE_ELSE;
							}
							break;
						}
						case 'n':
						{
							if (vala_scanner_matches (self, begin, "enum")) {
								return VALA_TOKEN_TYPE_ENUM;
							}
							break;
						}
					}
					break;
				}
				case 'l':
				{
					if (vala_scanner_matches (self, begin, "lock")) {
						return VALA_TOKEN_TYPE_LOCK;
					}
					break;
				}
				case 'n':
				{
					if (vala_scanner_matches (self, begin, "null")) {
						return VALA_TOKEN_TYPE_NULL;
					}
					break;
				}
				case 't':
				{
					switch (begin[1]) {
						case 'h':
						{
							if (vala_scanner_matches (self, begin, "this")) {
								return VALA_TOKEN_TYPE_THIS;
							}
							break;
						}
						case 'r':
						{
							if (vala_scanner_matches (self, begin, "true")) {
								return VALA_TOKEN_TYPE_TRUE;
							}
							break;
						}
					}
					break;
				}
				case 'v':
				{
					if (vala_scanner_matches (self, begin, "void")) {
						return VALA_TOKEN_TYPE_VOID;
					}
					break;
				}
				case 'w':
				{
					if (vala_scanner_matches (self, begin, "weak")) {
						return VALA_TOKEN_TYPE_WEAK;
					}
					break;
				}
			}
			break;
		}
		case 5:
		{
			switch (begin[0]) {
				case 'b':
				{
					if (vala_scanner_matches (self, begin, "break")) {
						return VALA_TOKEN_TYPE_BREAK;
					}
					break;
				}
				case 'c':
				{
					switch (begin[1]) {
						case 'a':
						{
							if (vala_scanner_matches (self, begin, "catch")) {
								return VALA_TOKEN_TYPE_CATCH;
							}
							break;
						}
						case 'l':
						{
							if (vala_scanner_matches (self, begin, "class")) {
								return VALA_TOKEN_TYPE_CLASS;
							}
							break;
						}
						case 'o':
						{
							if (vala_scanner_matches (self, begin, "const")) {
								return VALA_TOKEN_TYPE_CONST;
							}
							break;
						}
					}
					break;
				}
				case 'f':
				{
					if (vala_scanner_matches (self, begin, "false")) {
						return VALA_TOKEN_TYPE_FALSE;
					}
					break;
				}
				case 'o':
				{
					if (vala_scanner_matches (self, begin, "owned")) {
						return VALA_TOKEN_TYPE_OWNED;
					}
					break;
				}
				case 't':
				{
					if (vala_scanner_matches (self, begin, "throw")) {
						return VALA_TOKEN_TYPE_THROW;
					}
					break;
				}
				case 'u':
				{
					if (vala_scanner_matches (self, begin, "using")) {
						return VALA_TOKEN_TYPE_USING;
					}
					break;
				}
				case 'w':
				{
					if (vala_scanner_matches (self, begin, "while")) {
						return VALA_TOKEN_TYPE_WHILE;
					}
					break;
				}
				case 'y':
				{
					if (vala_scanner_matches (self, begin, "yield")) {
						return VALA_TOKEN_TYPE_YIELD;
					}
					break;
				}
			}
			break;
		}
		case 6:
		{
			switch (begin[0]) {
				case 'd':
				{
					if (vala_scanner_matches (self, begin, "delete")) {
						return VALA_TOKEN_TYPE_DELETE;
					}
					break;
				}
				case 'e':
				{
					if (vala_scanner_matches (self, begin, "extern")) {
						return VALA_TOKEN_TYPE_EXTERN;
					}
					break;
				}
				case 'i':
				{
					if (vala_scanner_matches (self, begin, "inline")) {
						return VALA_TOKEN_TYPE_INLINE;
					}
					break;
				}
				case 'p':
				{
					switch (begin[1]) {
						case 'a':
						{
							if (vala_scanner_matches (self, begin, "params")) {
								return VALA_TOKEN_TYPE_PARAMS;
							}
							break;
						}
						case 'u':
						{
							if (vala_scanner_matches (self, begin, "public")) {
								return VALA_TOKEN_TYPE_PUBLIC;
							}
							break;
						}
					}
					break;
				}
				case 'r':
				{
					if (vala_scanner_matches (self, begin, "return")) {
						return VALA_TOKEN_TYPE_RETURN;
					}
					break;
				}
				case 's':
				{
					switch (begin[1]) {
						case 'i':
						{
							switch (begin[2]) {
								case 'g':
								{
									if (vala_scanner_matches (self, begin, "signal")) {
										return VALA_TOKEN_TYPE_SIGNAL;
									}
									break;
								}
								case 'z':
								{
									if (vala_scanner_matches (self, begin, "sizeof")) {
										return VALA_TOKEN_TYPE_SIZEOF;
									}
									break;
								}
							}
							break;
						}
						case 't':
						{
							switch (begin[2]) {
								case 'a':
								{
									if (vala_scanner_matches (self, begin, "static")) {
										return VALA_TOKEN_TYPE_STATIC;
									}
									break;
								}
								case 'r':
								{
									if (vala_scanner_matches (self, begin, "struct")) {
										return VALA_TOKEN_TYPE_STRUCT;
									}
									break;
								}
							}
							break;
						}
						case 'w':
						{
							if (vala_scanner_matches (self, begin, "switch")) {
								return VALA_TOKEN_TYPE_SWITCH;
							}
							break;
						}
					}
					break;
				}
				case 't':
				{
					switch (begin[1]) {
						case 'h':
						{
							if (vala_scanner_matches (self, begin, "throws")) {
								return VALA_TOKEN_TYPE_THROWS;
							}
							break;
						}
						case 'y':
						{
							if (vala_scanner_matches (self, begin, "typeof")) {
								return VALA_TOKEN_TYPE_TYPEOF;
							}
							break;
						}
					}
					break;
				}
				case 'y':
				{
					if (vala_scanner_matches (self, begin, "yields")) {
						return VALA_TOKEN_TYPE_YIELDS;
					}
					break;
				}
			}
			break;
		}
		case 7:
		{
			switch (begin[0]) {
				case 'd':
				{
					switch (begin[1]) {
						case 'e':
						{
							if (vala_scanner_matches (self, begin, "default")) {
								return VALA_TOKEN_TYPE_DEFAULT;
							}
							break;
						}
						case 'y':
						{
							if (vala_scanner_matches (self, begin, "dynamic")) {
								return VALA_TOKEN_TYPE_DYNAMIC;
							}
							break;
						}
					}
					break;
				}
				case 'e':
				{
					if (vala_scanner_matches (self, begin, "ensures")) {
						return VALA_TOKEN_TYPE_ENSURES;
					}
					break;
				}
				case 'f':
				{
					switch (begin[1]) {
						case 'i':
						{
							if (vala_scanner_matches (self, begin, "finally")) {
								return VALA_TOKEN_TYPE_FINALLY;
							}
							break;
						}
						case 'o':
						{
							if (vala_scanner_matches (self, begin, "foreach")) {
								return VALA_TOKEN_TYPE_FOREACH;
							}
							break;
						}
					}
					break;
				}
				case 'p':
				{
					if (vala_scanner_matches (self, begin, "private")) {
						return VALA_TOKEN_TYPE_PRIVATE;
					}
					break;
				}
				case 'u':
				{
					if (vala_scanner_matches (self, begin, "unowned")) {
						return VALA_TOKEN_TYPE_UNOWNED;
					}
					break;
				}
				case 'v':
				{
					if (vala_scanner_matches (self, begin, "virtual")) {
						return VALA_TOKEN_TYPE_VIRTUAL;
					}
					break;
				}
			}
			break;
		}
		case 8:
		{
			switch (begin[0]) {
				case 'a':
				{
					if (vala_scanner_matches (self, begin, "abstract")) {
						return VALA_TOKEN_TYPE_ABSTRACT;
					}
					break;
				}
				case 'c':
				{
					if (vala_scanner_matches (self, begin, "continue")) {
						return VALA_TOKEN_TYPE_CONTINUE;
					}
					break;
				}
				case 'd':
				{
					if (vala_scanner_matches (self, begin, "delegate")) {
						return VALA_TOKEN_TYPE_DELEGATE;
					}
					break;
				}
				case 'i':
				{
					if (vala_scanner_matches (self, begin, "internal")) {
						return VALA_TOKEN_TYPE_INTERNAL;
					}
					break;
				}
				case 'o':
				{
					if (vala_scanner_matches (self, begin, "override")) {
						return VALA_TOKEN_TYPE_OVERRIDE;
					}
					break;
				}
				case 'r':
				{
					if (vala_scanner_matches (self, begin, "requires")) {
						return VALA_TOKEN_TYPE_REQUIRES;
					}
					break;
				}
				case 'v':
				{
					if (vala_scanner_matches (self, begin, "volatile")) {
						return VALA_TOKEN_TYPE_VOLATILE;
					}
					break;
				}
			}
			break;
		}
		case 9:
		{
			switch (begin[0]) {
				case 'c':
				{
					if (vala_scanner_matches (self, begin, "construct")) {
						return VALA_TOKEN_TYPE_CONSTRUCT;
					}
					break;
				}
				case 'i':
				{
					if (vala_scanner_matches (self, begin, "interface")) {
						return VALA_TOKEN_TYPE_INTERFACE;
					}
					break;
				}
				case 'n':
				{
					if (vala_scanner_matches (self, begin, "namespace")) {
						return VALA_TOKEN_TYPE_NAMESPACE;
					}
					break;
				}
				case 'p':
				{
					if (vala_scanner_matches (self, begin, "protected")) {
						return VALA_TOKEN_TYPE_PROTECTED;
					}
					break;
				}
			}
			break;
		}
		case 11:
		{
			if (vala_scanner_matches (self, begin, "errordomain")) {
				return VALA_TOKEN_TYPE_ERRORDOMAIN;
			}
			break;
		}
	}
	return VALA_TOKEN_TYPE_IDENTIFIER;
}


static ValaTokenType vala_scanner_read_number (ValaScanner* self) {
	ValaTokenType type;
	gboolean _tmp0;
	gboolean _tmp1;
	gboolean _tmp2;
	gboolean _tmp5;
	gboolean _tmp6;
	gboolean _tmp8;
	g_return_val_if_fail (self != NULL, 0);
	type = VALA_TOKEN_TYPE_INTEGER_LITERAL;
	_tmp0 = FALSE;
	_tmp1 = FALSE;
	_tmp2 = FALSE;
	if (self->priv->current < (self->priv->end - 2)) {
		_tmp2 = self->priv->current[0] == '0';
	} else {
		_tmp2 = FALSE;
	}
	if (_tmp2) {
		_tmp1 = self->priv->current[1] == 'x';
	} else {
		_tmp1 = FALSE;
	}
	if (_tmp1) {
		_tmp0 = g_ascii_isxdigit (self->priv->current[2]);
	} else {
		_tmp0 = FALSE;
	}
	/* integer part*/
	if (_tmp0) {
		/* hexadecimal integer literal*/
		self->priv->current = self->priv->current + 2;
		while (TRUE) {
			gboolean _tmp3;
			_tmp3 = FALSE;
			if (self->priv->current < self->priv->end) {
				_tmp3 = g_ascii_isxdigit (self->priv->current[0]);
			} else {
				_tmp3 = FALSE;
			}
			if (!_tmp3) {
				break;
			}
			self->priv->current++;
		}
	} else {
		while (TRUE) {
			gboolean _tmp4;
			_tmp4 = FALSE;
			if (self->priv->current < self->priv->end) {
				_tmp4 = g_ascii_isdigit (self->priv->current[0]);
			} else {
				_tmp4 = FALSE;
			}
			if (!_tmp4) {
				break;
			}
			self->priv->current++;
		}
	}
	_tmp5 = FALSE;
	_tmp6 = FALSE;
	if (self->priv->current < (self->priv->end - 1)) {
		_tmp6 = self->priv->current[0] == '.';
	} else {
		_tmp6 = FALSE;
	}
	if (_tmp6) {
		_tmp5 = g_ascii_isdigit (self->priv->current[1]);
	} else {
		_tmp5 = FALSE;
	}
	/* fractional part*/
	if (_tmp5) {
		type = VALA_TOKEN_TYPE_REAL_LITERAL;
		self->priv->current++;
		while (TRUE) {
			gboolean _tmp7;
			_tmp7 = FALSE;
			if (self->priv->current < self->priv->end) {
				_tmp7 = g_ascii_isdigit (self->priv->current[0]);
			} else {
				_tmp7 = FALSE;
			}
			if (!_tmp7) {
				break;
			}
			self->priv->current++;
		}
	}
	_tmp8 = FALSE;
	if (self->priv->current < self->priv->end) {
		_tmp8 = g_ascii_tolower (self->priv->current[0]) == 'e';
	} else {
		_tmp8 = FALSE;
	}
	/* exponent part*/
	if (_tmp8) {
		gboolean _tmp9;
		type = VALA_TOKEN_TYPE_REAL_LITERAL;
		self->priv->current++;
		_tmp9 = FALSE;
		if (self->priv->current < self->priv->end) {
			gboolean _tmp10;
			_tmp10 = FALSE;
			if (self->priv->current[0] == '+') {
				_tmp10 = TRUE;
			} else {
				_tmp10 = self->priv->current[0] == '-';
			}
			_tmp9 = _tmp10;
		} else {
			_tmp9 = FALSE;
		}
		if (_tmp9) {
			self->priv->current++;
		}
		while (TRUE) {
			gboolean _tmp11;
			_tmp11 = FALSE;
			if (self->priv->current < self->priv->end) {
				_tmp11 = g_ascii_isdigit (self->priv->current[0]);
			} else {
				_tmp11 = FALSE;
			}
			if (!_tmp11) {
				break;
			}
			self->priv->current++;
		}
	}
	/* type suffix*/
	if (self->priv->current < self->priv->end) {
		gboolean real_literal;
		gboolean _tmp15;
		real_literal = type == VALA_TOKEN_TYPE_REAL_LITERAL;
		switch (self->priv->current[0]) {
			case 'l':
			case 'L':
			{
				if (type == VALA_TOKEN_TYPE_INTEGER_LITERAL) {
					gboolean _tmp12;
					self->priv->current++;
					_tmp12 = FALSE;
					if (self->priv->current < self->priv->end) {
						_tmp12 = g_ascii_tolower (self->priv->current[0]) == 'l';
					} else {
						_tmp12 = FALSE;
					}
					if (_tmp12) {
						self->priv->current++;
					}
				}
				break;
			}
			case 'u':
			case 'U':
			{
				if (type == VALA_TOKEN_TYPE_INTEGER_LITERAL) {
					gboolean _tmp13;
					self->priv->current++;
					_tmp13 = FALSE;
					if (self->priv->current < self->priv->end) {
						_tmp13 = g_ascii_tolower (self->priv->current[0]) == 'l';
					} else {
						_tmp13 = FALSE;
					}
					if (_tmp13) {
						gboolean _tmp14;
						self->priv->current++;
						_tmp14 = FALSE;
						if (self->priv->current < self->priv->end) {
							_tmp14 = g_ascii_tolower (self->priv->current[0]) == 'l';
						} else {
							_tmp14 = FALSE;
						}
						if (_tmp14) {
							self->priv->current++;
						}
					}
				}
				break;
			}
			case 'f':
			case 'F':
			case 'd':
			case 'D':
			{
				type = VALA_TOKEN_TYPE_REAL_LITERAL;
				self->priv->current++;
				break;
			}
		}
		_tmp15 = FALSE;
		if (!real_literal) {
			_tmp15 = vala_scanner_is_ident_char (self, self->priv->current[0]);
		} else {
			_tmp15 = FALSE;
		}
		if (_tmp15) {
			while (TRUE) {
				gboolean _tmp16;
				_tmp16 = FALSE;
				if (self->priv->current < self->priv->end) {
					_tmp16 = vala_scanner_is_ident_char (self, self->priv->current[0]);
				} else {
					_tmp16 = FALSE;
				}
				if (!_tmp16) {
					break;
				}
				self->priv->current++;
			}
			type = VALA_TOKEN_TYPE_IDENTIFIER;
		}
	}
	return type;
}


ValaTokenType vala_scanner_read_token (ValaScanner* self, ValaSourceLocation* token_begin, ValaSourceLocation* token_end) {
	ValaTokenType type;
	gchar* begin;
	gint token_length_in_chars;
	g_return_val_if_fail (self != NULL, 0);
	vala_scanner_space (self);
	type = 0;
	begin = self->priv->current;
	(*token_begin).pos = begin;
	(*token_begin).line = self->priv->line;
	(*token_begin).column = self->priv->column;
	token_length_in_chars = -1;
	if (self->priv->current >= self->priv->end) {
		type = VALA_TOKEN_TYPE_EOF;
	} else {
		gboolean _tmp0;
		_tmp0 = FALSE;
		if (g_ascii_isalpha (self->priv->current[0])) {
			_tmp0 = TRUE;
		} else {
			_tmp0 = self->priv->current[0] == '_';
		}
		if (_tmp0) {
			gint len;
			len = 0;
			while (TRUE) {
				gboolean _tmp1;
				_tmp1 = FALSE;
				if (self->priv->current < self->priv->end) {
					_tmp1 = vala_scanner_is_ident_char (self, self->priv->current[0]);
				} else {
					_tmp1 = FALSE;
				}
				if (!_tmp1) {
					break;
				}
				self->priv->current++;
				len++;
			}
			type = vala_scanner_get_identifier_or_keyword (self, begin, len);
		} else {
			if (self->priv->current[0] == '@') {
				gint len;
				(*token_begin).pos++;
				/* @ is not part of the identifier*/
				self->priv->current++;
				len = 0;
				while (TRUE) {
					gboolean _tmp2;
					_tmp2 = FALSE;
					if (self->priv->current < self->priv->end) {
						_tmp2 = vala_scanner_is_ident_char (self, self->priv->current[0]);
					} else {
						_tmp2 = FALSE;
					}
					if (!_tmp2) {
						break;
					}
					self->priv->current++;
					len++;
				}
				type = VALA_TOKEN_TYPE_IDENTIFIER;
			} else {
				if (g_ascii_isdigit (self->priv->current[0])) {
					type = vala_scanner_read_number (self);
				} else {
					switch (self->priv->current[0]) {
						case '{':
						{
							type = VALA_TOKEN_TYPE_OPEN_BRACE;
							self->priv->current++;
							break;
						}
						case '}':
						{
							type = VALA_TOKEN_TYPE_CLOSE_BRACE;
							self->priv->current++;
							break;
						}
						case '(':
						{
							type = VALA_TOKEN_TYPE_OPEN_PARENS;
							self->priv->current++;
							break;
						}
						case ')':
						{
							type = VALA_TOKEN_TYPE_CLOSE_PARENS;
							self->priv->current++;
							break;
						}
						case '[':
						{
							type = VALA_TOKEN_TYPE_OPEN_BRACKET;
							self->priv->current++;
							break;
						}
						case ']':
						{
							type = VALA_TOKEN_TYPE_CLOSE_BRACKET;
							self->priv->current++;
							break;
						}
						case '.':
						{
							type = VALA_TOKEN_TYPE_DOT;
							self->priv->current++;
							if (self->priv->current < (self->priv->end - 1)) {
								gboolean _tmp3;
								_tmp3 = FALSE;
								if (self->priv->current[0] == '.') {
									_tmp3 = self->priv->current[1] == '.';
								} else {
									_tmp3 = FALSE;
								}
								if (_tmp3) {
									type = VALA_TOKEN_TYPE_ELLIPSIS;
									self->priv->current = self->priv->current + 2;
								}
							}
							break;
						}
						case ':':
						{
							gboolean _tmp4;
							type = VALA_TOKEN_TYPE_COLON;
							self->priv->current++;
							_tmp4 = FALSE;
							if (self->priv->current < self->priv->end) {
								_tmp4 = self->priv->current[0] == ':';
							} else {
								_tmp4 = FALSE;
							}
							if (_tmp4) {
								type = VALA_TOKEN_TYPE_DOUBLE_COLON;
								self->priv->current++;
							}
							break;
						}
						case ',':
						{
							type = VALA_TOKEN_TYPE_COMMA;
							self->priv->current++;
							break;
						}
						case ';':
						{
							type = VALA_TOKEN_TYPE_SEMICOLON;
							self->priv->current++;
							break;
						}
						case '#':
						{
							type = VALA_TOKEN_TYPE_HASH;
							self->priv->current++;
							break;
						}
						case '?':
						{
							type = VALA_TOKEN_TYPE_INTERR;
							self->priv->current++;
							break;
						}
						case '|':
						{
							type = VALA_TOKEN_TYPE_BITWISE_OR;
							self->priv->current++;
							if (self->priv->current < self->priv->end) {
								switch (self->priv->current[0]) {
									case '=':
									{
										type = VALA_TOKEN_TYPE_ASSIGN_BITWISE_OR;
										self->priv->current++;
										break;
									}
									case '|':
									{
										type = VALA_TOKEN_TYPE_OP_OR;
										self->priv->current++;
										break;
									}
								}
							}
							break;
						}
						case '&':
						{
							type = VALA_TOKEN_TYPE_BITWISE_AND;
							self->priv->current++;
							if (self->priv->current < self->priv->end) {
								switch (self->priv->current[0]) {
									case '=':
									{
										type = VALA_TOKEN_TYPE_ASSIGN_BITWISE_AND;
										self->priv->current++;
										break;
									}
									case '&':
									{
										type = VALA_TOKEN_TYPE_OP_AND;
										self->priv->current++;
										break;
									}
								}
							}
							break;
						}
						case '^':
						{
							gboolean _tmp5;
							type = VALA_TOKEN_TYPE_CARRET;
							self->priv->current++;
							_tmp5 = FALSE;
							if (self->priv->current < self->priv->end) {
								_tmp5 = self->priv->current[0] == '=';
							} else {
								_tmp5 = FALSE;
							}
							if (_tmp5) {
								type = VALA_TOKEN_TYPE_ASSIGN_BITWISE_XOR;
								self->priv->current++;
							}
							break;
						}
						case '~':
						{
							type = VALA_TOKEN_TYPE_TILDE;
							self->priv->current++;
							break;
						}
						case '=':
						{
							type = VALA_TOKEN_TYPE_ASSIGN;
							self->priv->current++;
							if (self->priv->current < self->priv->end) {
								switch (self->priv->current[0]) {
									case '=':
									{
										type = VALA_TOKEN_TYPE_OP_EQ;
										self->priv->current++;
										break;
									}
									case '>':
									{
										type = VALA_TOKEN_TYPE_LAMBDA;
										self->priv->current++;
										break;
									}
								}
							}
							break;
						}
						case '<':
						{
							type = VALA_TOKEN_TYPE_OP_LT;
							self->priv->current++;
							if (self->priv->current < self->priv->end) {
								switch (self->priv->current[0]) {
									case '=':
									{
										type = VALA_TOKEN_TYPE_OP_LE;
										self->priv->current++;
										break;
									}
									case '<':
									{
										gboolean _tmp6;
										type = VALA_TOKEN_TYPE_OP_SHIFT_LEFT;
										self->priv->current++;
										_tmp6 = FALSE;
										if (self->priv->current < self->priv->end) {
											_tmp6 = self->priv->current[0] == '=';
										} else {
											_tmp6 = FALSE;
										}
										if (_tmp6) {
											type = VALA_TOKEN_TYPE_ASSIGN_SHIFT_LEFT;
											self->priv->current++;
										}
										break;
									}
								}
							}
							break;
						}
						case '>':
						{
							gboolean _tmp7;
							type = VALA_TOKEN_TYPE_OP_GT;
							self->priv->current++;
							_tmp7 = FALSE;
							if (self->priv->current < self->priv->end) {
								_tmp7 = self->priv->current[0] == '=';
							} else {
								_tmp7 = FALSE;
							}
							if (_tmp7) {
								type = VALA_TOKEN_TYPE_OP_GE;
								self->priv->current++;
							}
							break;
						}
						case '!':
						{
							gboolean _tmp8;
							type = VALA_TOKEN_TYPE_OP_NEG;
							self->priv->current++;
							_tmp8 = FALSE;
							if (self->priv->current < self->priv->end) {
								_tmp8 = self->priv->current[0] == '=';
							} else {
								_tmp8 = FALSE;
							}
							if (_tmp8) {
								type = VALA_TOKEN_TYPE_OP_NE;
								self->priv->current++;
							}
							break;
						}
						case '+':
						{
							type = VALA_TOKEN_TYPE_PLUS;
							self->priv->current++;
							if (self->priv->current < self->priv->end) {
								switch (self->priv->current[0]) {
									case '=':
									{
										type = VALA_TOKEN_TYPE_ASSIGN_ADD;
										self->priv->current++;
										break;
									}
									case '+':
									{
										type = VALA_TOKEN_TYPE_OP_INC;
										self->priv->current++;
										break;
									}
								}
							}
							break;
						}
						case '-':
						{
							type = VALA_TOKEN_TYPE_MINUS;
							self->priv->current++;
							if (self->priv->current < self->priv->end) {
								switch (self->priv->current[0]) {
									case '=':
									{
										type = VALA_TOKEN_TYPE_ASSIGN_SUB;
										self->priv->current++;
										break;
									}
									case '-':
									{
										type = VALA_TOKEN_TYPE_OP_DEC;
										self->priv->current++;
										break;
									}
									case '>':
									{
										type = VALA_TOKEN_TYPE_OP_PTR;
										self->priv->current++;
										break;
									}
								}
							}
							break;
						}
						case '*':
						{
							gboolean _tmp9;
							type = VALA_TOKEN_TYPE_STAR;
							self->priv->current++;
							_tmp9 = FALSE;
							if (self->priv->current < self->priv->end) {
								_tmp9 = self->priv->current[0] == '=';
							} else {
								_tmp9 = FALSE;
							}
							if (_tmp9) {
								type = VALA_TOKEN_TYPE_ASSIGN_MUL;
								self->priv->current++;
							}
							break;
						}
						case '/':
						{
							gboolean _tmp10;
							type = VALA_TOKEN_TYPE_DIV;
							self->priv->current++;
							_tmp10 = FALSE;
							if (self->priv->current < self->priv->end) {
								_tmp10 = self->priv->current[0] == '=';
							} else {
								_tmp10 = FALSE;
							}
							if (_tmp10) {
								type = VALA_TOKEN_TYPE_ASSIGN_DIV;
								self->priv->current++;
							}
							break;
						}
						case '%':
						{
							gboolean _tmp11;
							type = VALA_TOKEN_TYPE_PERCENT;
							self->priv->current++;
							_tmp11 = FALSE;
							if (self->priv->current < self->priv->end) {
								_tmp11 = self->priv->current[0] == '=';
							} else {
								_tmp11 = FALSE;
							}
							if (_tmp11) {
								type = VALA_TOKEN_TYPE_ASSIGN_PERCENT;
								self->priv->current++;
							}
							break;
						}
						case '\'':
						case '"':
						{
							gboolean _tmp24;
							if (begin[0] == '\'') {
								type = VALA_TOKEN_TYPE_CHARACTER_LITERAL;
							} else {
								gboolean _tmp12;
								gboolean _tmp13;
								_tmp12 = FALSE;
								_tmp13 = FALSE;
								if (self->priv->current < (self->priv->end - 6)) {
									_tmp13 = begin[1] == '"';
								} else {
									_tmp13 = FALSE;
								}
								if (_tmp13) {
									_tmp12 = begin[2] == '"';
								} else {
									_tmp12 = FALSE;
								}
								if (_tmp12) {
									gboolean _tmp17;
									gboolean _tmp18;
									type = VALA_TOKEN_TYPE_VERBATIM_STRING_LITERAL;
									token_length_in_chars = 6;
									self->priv->current = self->priv->current + 3;
									while (self->priv->current < (self->priv->end - 4)) {
										gboolean _tmp14;
										gboolean _tmp15;
										_tmp14 = FALSE;
										_tmp15 = FALSE;
										if (self->priv->current[0] == '"') {
											_tmp15 = self->priv->current[1] == '"';
										} else {
											_tmp15 = FALSE;
										}
										if (_tmp15) {
											_tmp14 = self->priv->current[2] == '"';
										} else {
											_tmp14 = FALSE;
										}
										if (_tmp14) {
											break;
										} else {
											if (self->priv->current[0] == '\n') {
												self->priv->current++;
												self->priv->line++;
												self->priv->column = 1;
												token_length_in_chars = 3;
											} else {
												gunichar u;
												u = g_utf8_get_char_validated ((const char*) self->priv->current, (glong) (self->priv->end - self->priv->current));
												if (u != ((gunichar) (-1))) {
													self->priv->current = self->priv->current + g_unichar_to_utf8 (u, NULL);
													token_length_in_chars++;
												} else {
													ValaSourceReference* _tmp16;
													_tmp16 = NULL;
													vala_report_error (_tmp16 = vala_source_reference_new (self->priv->_source_file, self->priv->line, self->priv->column + token_length_in_chars, self->priv->line, self->priv->column + token_length_in_chars), "invalid UTF-8 character");
													(_tmp16 == NULL) ? NULL : (_tmp16 = (vala_source_reference_unref (_tmp16), NULL));
												}
											}
										}
									}
									_tmp17 = FALSE;
									_tmp18 = FALSE;
									if (self->priv->current[0] == '"') {
										_tmp18 = self->priv->current[1] == '"';
									} else {
										_tmp18 = FALSE;
									}
									if (_tmp18) {
										_tmp17 = self->priv->current[2] == '"';
									} else {
										_tmp17 = FALSE;
									}
									if (_tmp17) {
										self->priv->current = self->priv->current + 3;
									} else {
										ValaSourceReference* _tmp19;
										_tmp19 = NULL;
										vala_report_error (_tmp19 = vala_source_reference_new (self->priv->_source_file, self->priv->line, self->priv->column + token_length_in_chars, self->priv->line, self->priv->column + token_length_in_chars), "syntax error, expected \"\"\"");
										(_tmp19 == NULL) ? NULL : (_tmp19 = (vala_source_reference_unref (_tmp19), NULL));
									}
									break;
								} else {
									type = VALA_TOKEN_TYPE_STRING_LITERAL;
								}
							}
							token_length_in_chars = 2;
							self->priv->current++;
							while (TRUE) {
								gboolean _tmp20;
								_tmp20 = FALSE;
								if (self->priv->current < self->priv->end) {
									_tmp20 = self->priv->current[0] != begin[0];
								} else {
									_tmp20 = FALSE;
								}
								if (!_tmp20) {
									break;
								}
								if (self->priv->current[0] == '\\') {
									gboolean _tmp21;
									self->priv->current++;
									token_length_in_chars++;
									_tmp21 = FALSE;
									if (self->priv->current < self->priv->end) {
										_tmp21 = self->priv->current[0] == 'x';
									} else {
										_tmp21 = FALSE;
									}
									if (_tmp21) {
										/* hexadecimal escape character*/
										self->priv->current++;
										token_length_in_chars++;
										while (TRUE) {
											gboolean _tmp22;
											_tmp22 = FALSE;
											if (self->priv->current < self->priv->end) {
												_tmp22 = g_ascii_isxdigit (self->priv->current[0]);
											} else {
												_tmp22 = FALSE;
											}
											if (!_tmp22) {
												break;
											}
											self->priv->current++;
											token_length_in_chars++;
										}
									} else {
										self->priv->current++;
										token_length_in_chars++;
									}
								} else {
									if (self->priv->current[0] == '\n') {
										break;
									} else {
										gunichar u;
										u = g_utf8_get_char_validated ((const char*) self->priv->current, (glong) (self->priv->end - self->priv->current));
										if (u != ((gunichar) (-1))) {
											self->priv->current = self->priv->current + g_unichar_to_utf8 (u, NULL);
											token_length_in_chars++;
										} else {
											ValaSourceReference* _tmp23;
											_tmp23 = NULL;
											vala_report_error (_tmp23 = vala_source_reference_new (self->priv->_source_file, self->priv->line, self->priv->column + token_length_in_chars, self->priv->line, self->priv->column + token_length_in_chars), "invalid UTF-8 character");
											(_tmp23 == NULL) ? NULL : (_tmp23 = (vala_source_reference_unref (_tmp23), NULL));
										}
									}
								}
							}
							_tmp24 = FALSE;
							if (self->priv->current < self->priv->end) {
								_tmp24 = self->priv->current[0] != '\n';
							} else {
								_tmp24 = FALSE;
							}
							if (_tmp24) {
								self->priv->current++;
							} else {
								char* _tmp26;
								ValaSourceReference* _tmp25;
								_tmp26 = NULL;
								_tmp25 = NULL;
								vala_report_error (_tmp25 = vala_source_reference_new (self->priv->_source_file, self->priv->line, self->priv->column + token_length_in_chars, self->priv->line, self->priv->column + token_length_in_chars), _tmp26 = g_strdup_printf ("syntax error, expected %c", (gint) begin[0]));
								_tmp26 = (g_free (_tmp26), NULL);
								(_tmp25 == NULL) ? NULL : (_tmp25 = (vala_source_reference_unref (_tmp25), NULL));
							}
							break;
						}
						default:
						{
							gunichar u;
							u = g_utf8_get_char_validated ((const char*) self->priv->current, (glong) (self->priv->end - self->priv->current));
							if (u != ((gunichar) (-1))) {
								ValaSourceReference* _tmp27;
								self->priv->current = self->priv->current + g_unichar_to_utf8 (u, NULL);
								_tmp27 = NULL;
								vala_report_error (_tmp27 = vala_source_reference_new (self->priv->_source_file, self->priv->line, self->priv->column, self->priv->line, self->priv->column), "syntax error, unexpected character");
								(_tmp27 == NULL) ? NULL : (_tmp27 = (vala_source_reference_unref (_tmp27), NULL));
							} else {
								ValaSourceReference* _tmp28;
								self->priv->current++;
								_tmp28 = NULL;
								vala_report_error (_tmp28 = vala_source_reference_new (self->priv->_source_file, self->priv->line, self->priv->column, self->priv->line, self->priv->column), "invalid UTF-8 character");
								(_tmp28 == NULL) ? NULL : (_tmp28 = (vala_source_reference_unref (_tmp28), NULL));
							}
							self->priv->column++;
							return vala_scanner_read_token (self, &(*token_begin), &(*token_end));
						}
					}
				}
			}
		}
	}
	if (token_length_in_chars < 0) {
		self->priv->column = self->priv->column + ((gint) (self->priv->current - begin));
	} else {
		self->priv->column = self->priv->column + token_length_in_chars;
	}
	(*token_end).pos = self->priv->current;
	(*token_end).line = self->priv->line;
	(*token_end).column = self->priv->column - 1;
	return type;
}


static gboolean vala_scanner_matches (ValaScanner* self, gchar* begin, const char* keyword) {
	gchar* keyword_array;
	glong len;
	g_return_val_if_fail (self != NULL, FALSE);
	g_return_val_if_fail (keyword != NULL, FALSE);
	keyword_array = keyword;
	len = g_utf8_strlen (keyword, -1);
	{
		gint i;
		i = 0;
		for (; i < len; i++) {
			if (begin[i] != keyword_array[i]) {
				return FALSE;
			}
		}
	}
	return TRUE;
}


static gboolean vala_scanner_whitespace (ValaScanner* self) {
	gboolean found;
	g_return_val_if_fail (self != NULL, FALSE);
	found = FALSE;
	while (TRUE) {
		gboolean _tmp0;
		_tmp0 = FALSE;
		if (self->priv->current < self->priv->end) {
			_tmp0 = g_ascii_isspace (self->priv->current[0]);
		} else {
			_tmp0 = FALSE;
		}
		if (!_tmp0) {
			break;
		}
		if (self->priv->current[0] == '\n') {
			self->priv->line++;
			self->priv->column = 0;
		}
		found = TRUE;
		self->priv->current++;
		self->priv->column++;
	}
	return found;
}


static gboolean vala_scanner_comment (ValaScanner* self) {
	gboolean _tmp0;
	gboolean _tmp1;
	g_return_val_if_fail (self != NULL, FALSE);
	_tmp0 = FALSE;
	_tmp1 = FALSE;
	if (self->priv->current > (self->priv->end - 2)) {
		_tmp1 = TRUE;
	} else {
		_tmp1 = self->priv->current[0] != '/';
	}
	if (_tmp1) {
		_tmp0 = TRUE;
	} else {
		gboolean _tmp2;
		_tmp2 = FALSE;
		if (self->priv->current[1] != '/') {
			_tmp2 = self->priv->current[1] != '*';
		} else {
			_tmp2 = FALSE;
		}
		_tmp0 = _tmp2;
	}
	if (_tmp0) {
		return FALSE;
	}
	if (self->priv->current[1] == '/') {
		gchar* begin;
		char* _tmp5;
		/* single-line comment*/
		self->priv->current = self->priv->current + 2;
		begin = self->priv->current;
		while (TRUE) {
			gboolean _tmp4;
			_tmp4 = FALSE;
			if (self->priv->current < self->priv->end) {
				_tmp4 = self->priv->current[0] != '\n';
			} else {
				_tmp4 = FALSE;
			}
			if (!_tmp4) {
				break;
			}
			self->priv->current++;
		}
		_tmp5 = NULL;
		vala_scanner_push_comment (self, _tmp5 = g_strndup ((const char*) begin, (gsize) ((glong) (self->priv->current - begin))), self->priv->line == 1);
		_tmp5 = (g_free (_tmp5), NULL);
	} else {
		gchar* begin;
		gint begin_line;
		char* _tmp10;
		/* delimited comment*/
		self->priv->current = self->priv->current + 2;
		begin = self->priv->current;
		begin_line = self->priv->line;
		while (TRUE) {
			gboolean _tmp6;
			_tmp6 = FALSE;
			if (self->priv->current < (self->priv->end - 1)) {
				gboolean _tmp7;
				_tmp7 = FALSE;
				if (self->priv->current[0] != '*') {
					_tmp7 = TRUE;
				} else {
					_tmp7 = self->priv->current[1] != '/';
				}
				_tmp6 = _tmp7;
			} else {
				_tmp6 = FALSE;
			}
			if (!_tmp6) {
				break;
			}
			if (self->priv->current[0] == '\n') {
				self->priv->line++;
				self->priv->column = 0;
			}
			self->priv->current++;
			self->priv->column++;
		}
		if (self->priv->current == (self->priv->end - 1)) {
			ValaSourceReference* _tmp8;
			_tmp8 = NULL;
			vala_report_error (_tmp8 = vala_source_reference_new (self->priv->_source_file, self->priv->line, self->priv->column, self->priv->line, self->priv->column), "syntax error, expected */");
			(_tmp8 == NULL) ? NULL : (_tmp8 = (vala_source_reference_unref (_tmp8), NULL));
			return TRUE;
		}
		_tmp10 = NULL;
		vala_scanner_push_comment (self, _tmp10 = g_strndup ((const char*) begin, (gsize) ((glong) (self->priv->current - begin))), begin_line == 1);
		_tmp10 = (g_free (_tmp10), NULL);
		self->priv->current = self->priv->current + 2;
		self->priv->column = self->priv->column + 2;
	}
	return TRUE;
}


static void vala_scanner_space (ValaScanner* self) {
	g_return_if_fail (self != NULL);
	while (TRUE) {
		gboolean _tmp0;
		_tmp0 = FALSE;
		if (vala_scanner_whitespace (self)) {
			_tmp0 = TRUE;
		} else {
			_tmp0 = vala_scanner_comment (self);
		}
		if (!_tmp0) {
			break;
		}
	}
}


static void vala_scanner_push_comment (ValaScanner* self, const char* comment_item, gboolean file_comment) {
	g_return_if_fail (self != NULL);
	g_return_if_fail (comment_item != NULL);
	if (self->priv->_comment == NULL) {
		char* _tmp1;
		const char* _tmp0;
		_tmp1 = NULL;
		_tmp0 = NULL;
		self->priv->_comment = (_tmp1 = (_tmp0 = comment_item, (_tmp0 == NULL) ? NULL : g_strdup (_tmp0)), self->priv->_comment = (g_free (self->priv->_comment), NULL), _tmp1);
	} else {
		char* _tmp2;
		_tmp2 = NULL;
		self->priv->_comment = (_tmp2 = g_strdup_printf ("%s\n%s", self->priv->_comment, comment_item), self->priv->_comment = (g_free (self->priv->_comment), NULL), _tmp2);
	}
	if (file_comment) {
		char* _tmp3;
		vala_source_file_set_comment (self->priv->_source_file, self->priv->_comment);
		_tmp3 = NULL;
		self->priv->_comment = (_tmp3 = NULL, self->priv->_comment = (g_free (self->priv->_comment), NULL), _tmp3);
	}
}


/**
 * Clears and returns the content of the comment stack.
 *
 * @return saved comment
 */
char* vala_scanner_pop_comment (ValaScanner* self) {
	GString* result_builder;
	char* _tmp1;
	const char* index;
	const char* _tmp2;
	char* _tmp3;
	g_return_val_if_fail (self != NULL, NULL);
	if (self->priv->_comment == NULL) {
		return NULL;
	}
	result_builder = g_string_new (self->priv->_comment);
	_tmp1 = NULL;
	self->priv->_comment = (_tmp1 = NULL, self->priv->_comment = (g_free (self->priv->_comment), NULL), _tmp1);
	index = NULL;
	while ((index = g_utf8_strchr (result_builder->str, (glong) (-1), (gunichar) '\t')) != NULL) {
		g_string_erase (result_builder, g_utf8_pointer_to_offset (result_builder->str, index), (glong) 1);
	}
	_tmp2 = NULL;
	_tmp3 = NULL;
	return (_tmp3 = (_tmp2 = result_builder->str, (_tmp2 == NULL) ? NULL : g_strdup (_tmp2)), (result_builder == NULL) ? NULL : (result_builder = (g_string_free (result_builder, TRUE), NULL)), _tmp3);
}


ValaSourceFile* vala_scanner_get_source_file (ValaScanner* self) {
	g_return_val_if_fail (self != NULL, NULL);
	return self->priv->_source_file;
}


static void vala_scanner_set_source_file (ValaScanner* self, ValaSourceFile* value) {
	ValaSourceFile* _tmp2;
	ValaSourceFile* _tmp1;
	g_return_if_fail (self != NULL);
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_source_file = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL) ? NULL : vala_source_file_ref (_tmp1)), (self->priv->_source_file == NULL) ? NULL : (self->priv->_source_file = (vala_source_file_unref (self->priv->_source_file), NULL)), _tmp2);
}


static void vala_value_scanner_init (GValue* value) {
	value->data[0].v_pointer = NULL;
}


static void vala_value_scanner_free_value (GValue* value) {
	if (value->data[0].v_pointer) {
		vala_scanner_unref (value->data[0].v_pointer);
	}
}


static void vala_value_scanner_copy_value (const GValue* src_value, GValue* dest_value) {
	if (src_value->data[0].v_pointer) {
		dest_value->data[0].v_pointer = vala_scanner_ref (src_value->data[0].v_pointer);
	} else {
		dest_value->data[0].v_pointer = NULL;
	}
}


static gpointer vala_value_scanner_peek_pointer (const GValue* value) {
	return value->data[0].v_pointer;
}


static gchar* vala_value_scanner_collect_value (GValue* value, guint n_collect_values, GTypeCValue* collect_values, guint collect_flags) {
	if (collect_values[0].v_pointer) {
		ValaScanner* object;
		object = collect_values[0].v_pointer;
		if (object->parent_instance.g_class == NULL) {
			return g_strconcat ("invalid unclassed object pointer for value type `", G_VALUE_TYPE_NAME (value), "'", NULL);
		} else if (!g_value_type_compatible (G_TYPE_FROM_INSTANCE (object), G_VALUE_TYPE (value))) {
			return g_strconcat ("invalid object type `", g_type_name (G_TYPE_FROM_INSTANCE (object)), "' for value type `", G_VALUE_TYPE_NAME (value), "'", NULL);
		}
		value->data[0].v_pointer = vala_scanner_ref (object);
	} else {
		value->data[0].v_pointer = NULL;
	}
	return NULL;
}


static gchar* vala_value_scanner_lcopy_value (const GValue* value, guint n_collect_values, GTypeCValue* collect_values, guint collect_flags) {
	ValaScanner** object_p;
	object_p = collect_values[0].v_pointer;
	if (!object_p) {
		return g_strdup_printf ("value location for `%s' passed as NULL", G_VALUE_TYPE_NAME (value));
	}
	if (!value->data[0].v_pointer) {
		*object_p = NULL;
	} else if (collect_flags && G_VALUE_NOCOPY_CONTENTS) {
		*object_p = value->data[0].v_pointer;
	} else {
		*object_p = vala_scanner_ref (value->data[0].v_pointer);
	}
	return NULL;
}


GParamSpec* vala_param_spec_scanner (const gchar* name, const gchar* nick, const gchar* blurb, GType object_type, GParamFlags flags) {
	ValaParamSpecScanner* spec;
	g_return_val_if_fail (g_type_is_a (object_type, VALA_TYPE_SCANNER), NULL);
	spec = g_param_spec_internal (G_TYPE_PARAM_OBJECT, name, nick, blurb, flags);
	G_PARAM_SPEC (spec)->value_type = object_type;
	return G_PARAM_SPEC (spec);
}


gpointer vala_value_get_scanner (const GValue* value) {
	g_return_val_if_fail (G_TYPE_CHECK_VALUE_TYPE (value, VALA_TYPE_SCANNER), NULL);
	return value->data[0].v_pointer;
}


void vala_value_set_scanner (GValue* value, gpointer v_object) {
	ValaScanner* old;
	g_return_if_fail (G_TYPE_CHECK_VALUE_TYPE (value, VALA_TYPE_SCANNER));
	old = value->data[0].v_pointer;
	if (v_object) {
		g_return_if_fail (G_TYPE_CHECK_INSTANCE_TYPE (v_object, VALA_TYPE_SCANNER));
		g_return_if_fail (g_value_type_compatible (G_TYPE_FROM_INSTANCE (v_object), G_VALUE_TYPE (value)));
		value->data[0].v_pointer = v_object;
		vala_scanner_ref (value->data[0].v_pointer);
	} else {
		value->data[0].v_pointer = NULL;
	}
	if (old) {
		vala_scanner_unref (old);
	}
}


static void vala_scanner_class_init (ValaScannerClass * klass) {
	vala_scanner_parent_class = g_type_class_peek_parent (klass);
	VALA_SCANNER_CLASS (klass)->finalize = vala_scanner_finalize;
	g_type_class_add_private (klass, sizeof (ValaScannerPrivate));
}


static void vala_scanner_instance_init (ValaScanner * self) {
	self->priv = VALA_SCANNER_GET_PRIVATE (self);
	self->ref_count = 1;
}


static void vala_scanner_finalize (ValaScanner* obj) {
	ValaScanner * self;
	self = VALA_SCANNER (obj);
	(self->priv->_source_file == NULL) ? NULL : (self->priv->_source_file = (vala_source_file_unref (self->priv->_source_file), NULL));
	self->priv->_comment = (g_free (self->priv->_comment), NULL);
}


GType vala_scanner_get_type (void) {
	static GType vala_scanner_type_id = 0;
	if (vala_scanner_type_id == 0) {
		static const GTypeValueTable g_define_type_value_table = { vala_value_scanner_init, vala_value_scanner_free_value, vala_value_scanner_copy_value, vala_value_scanner_peek_pointer, "p", vala_value_scanner_collect_value, "p", vala_value_scanner_lcopy_value };
		static const GTypeInfo g_define_type_info = { sizeof (ValaScannerClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) vala_scanner_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (ValaScanner), 0, (GInstanceInitFunc) vala_scanner_instance_init, &g_define_type_value_table };
		static const GTypeFundamentalInfo g_define_type_fundamental_info = { (G_TYPE_FLAG_CLASSED | G_TYPE_FLAG_INSTANTIATABLE | G_TYPE_FLAG_DERIVABLE | G_TYPE_FLAG_DEEP_DERIVABLE) };
		vala_scanner_type_id = g_type_register_fundamental (g_type_fundamental_next (), "ValaScanner", &g_define_type_info, &g_define_type_fundamental_info, 0);
	}
	return vala_scanner_type_id;
}


gpointer vala_scanner_ref (gpointer instance) {
	ValaScanner* self;
	self = instance;
	g_atomic_int_inc (&self->ref_count);
	return instance;
}


void vala_scanner_unref (gpointer instance) {
	ValaScanner* self;
	self = instance;
	if (g_atomic_int_dec_and_test (&self->ref_count)) {
		VALA_SCANNER_GET_CLASS (self)->finalize (self);
		g_type_free_instance ((GTypeInstance *) self);
	}
}




